@@ -83,6 +83,9 @@ Evo.foreachIFrameDocument = function(doc, traversar_obj, call_also_for_doc, leve
 	iframes = doc.getElementsByTagName("iframe");
 
 	for (ii = 0; ii < iframes.length; ii++) {
+		if (!iframes[ii].contentDocument)
+			continue;
+
 		if (!traversar_obj.exec(iframes[ii].contentDocument, iframes[ii].id, level + 1))
 			return false;
 
@@ -692,21 +695,6 @@ Evo.initialize = function(elem)
 
 	if (doc.defaultView && !doc.defaultView.frameElement && !doc.body.hasAttribute("class"))
 		doc.body.className = "-e-web-view-background-color -e-web-view-text-color";
-
-	if (doc.documentElement.style.getPropertyValue("color") == "" ||
-	    doc.documentElement.style.getPropertyValue("color") == "text") {
-		if (doc.defaultView && doc.defaultView.frameElement && !doc.defaultView.frameElement.hasAttribute("x-e-unset-colors") &&
-		    doc.defaultView.frameElement.ownerDocument &&
-		    doc.defaultView.frameElement.ownerDocument.defaultView &&
-		    doc.defaultView.frameElement.ownerDocument.defaultView.window) {
-			var style = doc.defaultView.frameElement.ownerDocument.defaultView.window.getComputedStyle(doc.defaultView.frameElement);
-			if (style) {
-				doc.documentElement.style.setProperty("color", style.color);
-				doc.documentElement.style.setProperty("background-color", style.backgroundColor);
-			}
-		}
-	}
-
 	elems = doc.querySelectorAll("input, textarea, select, button, label");
 
 	for (ii = 0; ii < elems.length; ii++) {
@@ -772,6 +760,41 @@ Evo.EnsureMainDocumentInitialized = function()
 	Evo.initializeAndPostContentLoaded(null);
 }
 
+Evo.mailDisplayUpdateIFramesHeightRecursive = function(doc)
+{
+	if (!doc)
+		return;
+
+	var ii, iframes;
+
+	iframes = doc.getElementsByTagName("iframe");
+
+	/* Update from bottom to top */
+	for (ii = 0; ii < iframes.length; ii++) {
+		Evo.mailDisplayUpdateIFramesHeightRecursive(iframes[ii].contentDocument);
+	}
+
+	if (!doc.scrollingElement || !doc.defaultView || !doc.defaultView.frameElement)
+		return;
+
+	if (doc.defaultView.frameElement.height == doc.scrollingElement.scrollHeight)
+		doc.defaultView.frameElement.height = 10;
+	doc.defaultView.frameElement.height = doc.scrollingElement.scrollHeight + 2 + (doc.scrollingElement.scrollWidth > doc.scrollingElement.clientWidth ? 20 : 0);
+}
+
+Evo.MailDisplayUpdateIFramesHeight = function()
+{
+	var scrolly = document.defaultView ? document.defaultView.scrollY : -1;
+
+	Evo.mailDisplayUpdateIFramesHeightRecursive(document);
+
+	if (scrolly != -1 && document.defaultView.scrollY != scrolly)
+		document.defaultView.scrollTo(0, scrolly);
+
+	Evo.mailDisplayResizeContentToPreviewWidth();
+	Evo.mailDisplayUpdateMagicSpacebarState();
+}
+
 if (this instanceof Window && this.document) {
 	this.document.onload = function() { Evo.initializeAndPostContentLoaded(this); };
 
@@ -857,9 +880,8 @@ Evo.mailDisplayResizeContentToPreviewWidth = function()
 				local_width -= 2; /* 1 + 1 frame borders */
 			} else if (!iframes.length) {
 				/* Message main body */
-				local_width -= 8; /* 8 + 8 margins of body without iframes */
-				if (level > 1)
-					local_width -= 8;
+				local_width -= level * 20; /* 10 + 10 margins of body without iframes */
+				local_width -= 4;
 
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", "body", "width: " + local_width + "px;");
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", ".part-container", "width: " + local_width + "px;");
@@ -869,7 +891,7 @@ Evo.mailDisplayResizeContentToPreviewWidth = function()
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", "body",
 					"width: " + local_width + "px;");
 
-				local_width -= 2; /* 1 + 1 frame borders */
+				local_width -= 4; /* 2 + 2 frame borders */
 
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", ".part-container-nostyle iframe",
 					"width: " + local_width + "px;");
@@ -881,30 +903,53 @@ Evo.mailDisplayResizeContentToPreviewWidth = function()
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", ".part-container iframe",
 					"width: " + (local_width - 10) + "px;");
 			} else {
-				local_width -= 20; /* 10 + 10 margins of body with iframes */
-				local_width -= 8; /* attachment margin */
-				local_width -= 2; /* 1 + 1 frame borders */
+				local_width -= (level - 1) * 20; /* 10 + 10 margins of body with iframes */
+				local_width -= 4; /* 2 + 2 frame borders */
+				local_width -= 10; /* attachment margin */
 
-				/* We need to subtract another 10 pixels from the iframe width to
-				 * have the iframe's borders on the correct place. We can't subtract
-				 * it from local_width as we don't want to propagate this change
-				 * further. */
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", ".part-container-nostyle iframe",
-					"width: " + (local_width - 10) + "px;");
+					"width: " + local_width + "px;");
 
 				Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", "body > .part-container-nostyle iframe",
-					"width: " + (local_width - 10) + "px;");
+					"width: " + local_width + "px;");
 			}
 
 			/* Add rules to every sub document */
 			for (ii = 0; ii < iframes.length; ii++) {
+				var jj, frmdoc = iframes[ii].contentDocument;
+				for (jj = 0; frmdoc && jj < frmdoc.images.length; jj++) {
+					var img = frmdoc.images[jj];
+					if (frmdoc.defaultView && !img.hasAttribute("width") && !img.hasAttribute("height")) {
+						var can1 = img.hasAttribute("x-evo-width-modified"), can2 = false;
+						if (!can1)
+							can2 = img.style.width == "" && img.style.height == "";
+						if (can1 || can2) {
+							var expected_width;
+							if (can1) {
+								expected_width = parseInt(img.getAttribute("x-evo-width-modified"));
+							} else {
+								var tmp = frmdoc.defaultView.getComputedStyle(img).width;
+								if (tmp && tmp.endsWith("px"))
+									expected_width = parseInt(tmp.slice(0, -2));
+								else
+									expected_width = tmp;
+								img.setAttribute("x-evo-width-modified", expected_width);
+							}
+							if (expected_width < local_width)
+								img.style.width = expected_width + "px";
+							else
+								img.style.width = local_width + "px";
+						}
+					}
+				}
+
 				if (!this.can_force_width_on_iframe (iframes[ii]))
 					continue;
 
 				var tmp_local_width = local_width;
 
 				if (level == 0) {
-					tmp_local_width -= 8; /* attachment's margin */
+					tmp_local_width -= 10; /* attachment's margin */
 
 					Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", ".attachment-wrapper iframe:not([src*=\"__formatas=\"])",
 						"width: " + tmp_local_width + "px;");
@@ -913,7 +958,7 @@ Evo.mailDisplayResizeContentToPreviewWidth = function()
 						"width: " + tmp_local_width + "px;");
 
 					Evo.addRuleIntoStyleSheetDocument(doc, "-e-mail-formatter-style-sheet", "body > .part-container-nostyle iframe",
-						"width: " + local_width + "px;");
+						"width: " + tmp_local_width + "px;");
 				}
 
 				this.set_iframe_and_body_width (iframes[ii].contentDocument, tmp_local_width, original_width, level + 1);
@@ -926,6 +971,9 @@ Evo.mailDisplayResizeContentToPreviewWidth = function()
 	width -= 20; /* 10 + 10 margins of body */
 
 	traversar.set_iframe_and_body_width(document, width, width, 0);
+
+	if (document.documentElement.clientWidth - 20 > width)
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 }
 
 Evo.mailDisplayUpdateMagicSpacebarState = function()
@@ -1253,8 +1301,10 @@ Evo.MailDisplayBindDOM = function(iframe_id, markCitationColor)
 	Evo.mailDisplayResizeContentToPreviewWidth();
 	Evo.mailDisplayUpdateMagicSpacebarState();
 
-	document.defaultView.onresize = Evo.mailDisplayResized;
-	document.defaultView.onscroll = Evo.mailDisplayUpdateMagicSpacebarState;
+	if (document.body) {
+		document.body.onresize = Evo.mailDisplayResized;
+		document.body.onscroll = Evo.mailDisplayUpdateMagicSpacebarState;
+	}
 }
 
 Evo.MailDisplayShowAttachment = function(element_id, show)
@@ -1294,6 +1344,8 @@ Evo.MailDisplayShowAttachment = function(element_id, show)
 			window.webkit.messageHandlers.contentLoaded.postMessage(iframe_id);
 			Evo.mailDisplayUpdateMagicSpacebarState();
 		}
+	} else if (elem.ownerDocument.defaultView.frameElement) {
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
@@ -1388,8 +1440,10 @@ EvoItip.SetElementInnerHTML = function(iframe_id, element_id, html_content)
 {
 	var elem = Evo.FindElement(iframe_id, element_id);
 
-	if (elem)
+	if (elem) {
 		elem.innerHTML = html_content;
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
+	}
 }
 
 EvoItip.SetShowCheckbox = function(iframe_id, element_id, show, update_second)
@@ -1415,6 +1469,8 @@ EvoItip.SetShowCheckbox = function(iframe_id, element_id, show, update_second)
 		if (elem) {
 			elem.hidden = !show;
 		}
+
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
@@ -1428,6 +1484,8 @@ EvoItip.SetAreaText = function(iframe_id, element_id, text)
 		if (row.lastElementChild) {
 			row.lastElementChild.innerHTML = text;
 		}
+
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
@@ -1445,6 +1503,8 @@ EvoItip.UpdateTimes = function(iframe_id, element_id, header, label)
 		if (elem.lastElementChild) {
 			elem.lastElementChild.innerHTML = label;
 		}
+
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
@@ -1472,6 +1532,8 @@ EvoItip.AppendInfoRow = function(iframe_id, table_id, row_id, icon_name, message
 
 	cell = row.insertCell(-1);
 	cell.innerHTML = message;
+
+	window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 }
 
 EvoItip.RemoveInfoRow = function(iframe_id, row_id)
@@ -1480,6 +1542,7 @@ EvoItip.RemoveInfoRow = function(iframe_id, row_id)
 
 	if (row && row.parentNode) {
 		row.parentNode.removeChild(row);
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
@@ -1491,6 +1554,8 @@ EvoItip.RemoveChildNodes = function(iframe_id, element_id)
 		while (elem.lastChild) {
 			elem.removeChild(elem.lastChild);
 		}
+
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
@@ -1543,6 +1608,8 @@ EvoItip.HideButtons = function(iframe_id, element_id)
 			if (button)
 				button.hidden = true;
 		}
+
+		window.webkit.messageHandlers.scheduleIFramesHeightUpdate.postMessage(0);
 	}
 }
 
